/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
\---------------------------------------------------------------------*/
// Automatically generated wrapper for zypp/Bit.h
// Source Namespace: zypp -> Target Namespace: zyppng
#ifndef BIT_H_ZYPPNG_WRAPPER_H
#define BIT_H_ZYPPNG_WRAPPER_H

#include <zypp/Bit.h>

namespace zyppng
{
    namespace bit {
        using ::zypp::bit::MaxBits;
        using ::zypp::bit::asString;
        using ::zypp::bit::Mask;
        using ::zypp::bit::Range;
        using ::zypp::bit::RangeValue;
        using ::zypp::bit::RangeBit;
        using ::zypp::bit::BitField;
        using ::zypp::bit::operator<<;
        using ::zypp::bit::operator==;
        using ::zypp::bit::operator!=;
        using ::zypp::bit::operator&;
        using ::zypp::bit::operator|;
        using ::zypp::bit::operator^;
        using ::zypp::bit::operator>>;
    } // namespace bit

} // namespace zyppng
#endif // BIT_H_ZYPPNG_WRAPPER_H
