/*---------------------------------------------------------------------\
|                          ____ _   __ __ ___                          |
|                         |__  / \ / / . \ . \                         |
|                           / / \ V /|  _/  _/                         |
|                          / /__ | | | | | |                           |
|                         /_____||_| |_| |_|                           |
|                                                                      |
\---------------------------------------------------------------------*/
// Automatically generated wrapper for Arch.h
// Source Namespace: zypp -> Target Namespace: zyppng
#ifndef ZYPPNG_ARCH_H_WRAPPER_H
#define ZYPPNG_ARCH_H_WRAPPER_H

#include <zypp/Arch.h>

namespace zyppng
{
  using ::zypp::Arch;
  using ::zypp::Arch_empty;
  using ::zypp::Arch_noarch;
  using ::zypp::Arch_pentium4;
  using ::zypp::Arch_pentium3;
  using ::zypp::Arch_x86_64_v4;
  using ::zypp::Arch_x86_64_v3;
  using ::zypp::Arch_x86_64_v2;
  using ::zypp::Arch_x86_64;
  using ::zypp::Arch_athlon;
  using ::zypp::Arch_i686;
  using ::zypp::Arch_i586;
  using ::zypp::Arch_i486;
  using ::zypp::Arch_i386;
  using ::zypp::Arch_s390x;
  using ::zypp::Arch_s390;
  using ::zypp::Arch_ppc64le;
  using ::zypp::Arch_ppc64p7;
  using ::zypp::Arch_ppc64;
  using ::zypp::Arch_ppc;
  using ::zypp::Arch_ia64;
  using ::zypp::Arch_alphaev67;
  using ::zypp::Arch_alphaev6;
  using ::zypp::Arch_alphapca56;
  using ::zypp::Arch_alphaev56;
  using ::zypp::Arch_alphaev5;
  using ::zypp::Arch_alpha;
  using ::zypp::Arch_sparc64v;
  using ::zypp::Arch_sparc64;
  using ::zypp::Arch_sparcv9v;
  using ::zypp::Arch_sparcv9;
  using ::zypp::Arch_sparcv8;
  using ::zypp::Arch_sparc;
  using ::zypp::Arch_aarch64;
  using ::zypp::Arch_armv7tnh;
  using ::zypp::Arch_armv7thl;
  using ::zypp::Arch_armv7hnl;
  using ::zypp::Arch_armv7hl;
  using ::zypp::Arch_armv6hl;
  using ::zypp::Arch_armv8hl;
  using ::zypp::Arch_armv8l;
  using ::zypp::Arch_armv7l;
  using ::zypp::Arch_armv6l;
  using ::zypp::Arch_armv5tejl;
  using ::zypp::Arch_armv5tel;
  using ::zypp::Arch_armv5tl;
  using ::zypp::Arch_armv5l;
  using ::zypp::Arch_armv4tl;
  using ::zypp::Arch_armv4l;
  using ::zypp::Arch_armv3l;
  using ::zypp::Arch_riscv64;
  using ::zypp::Arch_sh3;
  using ::zypp::Arch_sh4;
  using ::zypp::Arch_sh4a;
  using ::zypp::Arch_m68k;
  using ::zypp::Arch_mips;
  using ::zypp::Arch_mipsel;
  using ::zypp::Arch_mips64;
  using ::zypp::Arch_mips64el;
  using ::zypp::Arch_loong64;
  using ::zypp::operator<<;
  using ::zypp::dumpAsXmlOn;
  using ::zypp::operator==;
  using ::zypp::operator!=;
  using ::zypp::ArchCompatibleWith;
} // namespace zyppng
#endif // ZYPPNG_ARCH_H_WRAPPER_H
